@@ -42,7 +42,6 @@
 #include "camel-mapi-store.h"
 #include "camel-mapi-folder.h"
 #include "camel-mapi-store-summary.h"
-#include "camel-mapi-utils.h"
 #define d(x)
 
 #include <exchange-mapi-defs.h>
@@ -63,8 +62,18 @@ mapi_message_item_send (ExchangeMapiConnection *conn, MailItem *item, GError **p
 	guint64 fid = 0;
 	mapi_id_t mid = 0;
 
+	#define unset(x) g_free (x); x = NULL
+
+	item->header.flags = MSGFLAG_UNSENT;
+	unset (item->header.from);
+	unset (item->header.from_email);
+	unset (item->header.transport_headers);
+	item->header.recieved_time = 0;
+
+	#undef unset
+
 	mid = exchange_mapi_connection_create_item (conn, olFolderSentMail, fid,
-					 camel_mapi_utils_create_item_build_props, item,
+					 mapi_mail_utils_create_item_build_props, item,
 					 item->recipients, item->attachments, item->generic_streams, MAPI_OPTIONS_DELETE_ON_SUBMIT_FAILURE, perror);
 
 	return mid;
@@ -105,7 +114,7 @@ mapi_send_to_sync (CamelTransport *transport,
 	}
 
 	/* Convert MIME to MailItem, attacment lists and recipient list.*/
-	item = camel_mapi_utils_mime_to_item (message, from, cancellable, NULL);
+	item = mapi_mime_message_to_mail_item (message, 0, from, cancellable, NULL);
 
 	/* send */
 	st = mapi_message_item_send (conn, item, error);
